/*
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

#include "proton/tracker.hpp"

#include "proton/sender.hpp"

#include "proton_bits.hpp"
#include "types_internal.hpp"
#include "proton/binary.hpp"

#include <proton/delivery.h>

namespace proton {

tracker::tracker(pn_delivery_t *d): transfer(make_wrapper(d)) {}
sender tracker::sender() const { return make_wrapper<class sender>(pn_delivery_link(pn_object())); }
binary tracker::tag() const { return bin(pn_delivery_tag(pn_object())); }
}
